import .syntax
import .common

namespace imp
namespace interp

def op1 : unary_op → val → option val
| unary_op.neg (val.int i) := some (val.int (-i))
| unary_op.not (val.bool b) := some (val.bool $ bnot b)
| _ _ := none

def op2 : binary_op → val → val → option val
| binary_op.add (val.int i1) (val.int i2) :=
    some (val.int (i1 + i2))
| binary_op.add (val.str s1) (val.str s2) :=
    some (val.str (s1 ++ s2))
 /-
    5 points:

    TODO Complete the rest of interp_op2.

    Note: instead of `/`, `%` please use `int.fdiv`
    and `int.fmod` for defining division, and modulus
    respectively.
  -/
| _ _ _ := none

def exp (s : store) (h : heap) : expr → option val
| (expr.val v) := some v
| (expr.var x) := store.lookup x s
| (expr.unary_op op e1) :=
  do v1 ← exp e1,
      op1 op v1
| (expr.binary_op op e1 e2) :=
  do v1 ← exp e1,
     v2 ← exp e2,
     op2 op v1 v2
| (expr.length e1) :=
  do v1 ← exp e1,
     match v1 with
     | (val.addr a) :=
       match heap.read h a with
       | some (val.int i) := some (val.int i)
       | _ := none
       end
     | (val.str cs) :=
        some (val.int cs.length)
     | _ := none
     end
| (expr.index e1 e2) :=
  do v1 ← exp e1,
     v2 ← exp e2,
     match (v1, v2) with
     | ((val.addr a), (val.int i)) :=
       do v1 ← heap.read h a,
          match v1 with
          | (val.int l) :=
              if 0 <= i ∧ i < l
              then heap.read h (a + i + 1)
              else none
          | _ := none
          end
     | ((val.str cs), (val.int i)) :=
          if 0 <= i
          then do c ← cs.index i.to_nat,
                  some $ val.str (string.str "" c)
          else none
     | (_, _) := none
  end

def seq_exp (s : store) (h : heap) : list expr → option (list val)
| [] := some []
| (e :: t) :=
  do v ← exp s h e,
     vs ← seq_exp t,
     some (v :: vs)

end interp

def interp.stmt (ioi : io.interface) : nat → env → store → heap → stmt → option (store × heap)
| 0 _ _ _ _ := none
| (n + 1) e s h (stmt.alloc x e1 e2) :=
  do v1 ← interp.exp s h e1,
     v2 ← interp.exp s h e2,
     match (v1, v2) with
     | (val.int i, v) :=
      if 0 <= i
      then some (update x (val.addr h.length) s, heap.alloc h i v)
      else none
     | (_, _) := none
     end
| (n + 1) e s h (stmt.write x e1 e2) :=
  do v1 ← store.lookup x s,
     v2 ← interp.exp s h e1,
     v3 ← interp.exp s h e2,
     match (v1, v2, v3) with
     | (val.addr a, val.int i, v) :=
       do v4 ← heap.read h a,
          match v4 with
          | val.int l :=
            if 0 <= i ∧ i < l
            then match heap.write h (a + i + 1) v with
                    | some h' := some (s, h')
                    | none := none
                    end
            else none
          | _ := none
          end
      | (_, _, _) := none
      end
| (n + 1) e s h (stmt.call x f es) :=
  do vs ← interp.seq_exp s h es,
     match locate e f with
     | some (func.mk _ params body ret) :=
        do sf ← updates params vs store.empty,
           (s', h') ← interp.stmt n e sf h body,
           v' ← interp.exp s' h' ret,
           some (update x v' s, h')
     | none :=
      do (v', h') ← some $ extcall ioi f vs h,
         some (update x v' s, h')
     end
    -- (**
    --    10 points:

    --    TODO Complete the rest of interp_s.
    -- *)
| (n + 1) e s h stmt.nop := none
| (n + 1) ev s h (stmt.set x e) := none
| (n + 1) env s h (stmt.ifelse e p1 p2) :=
  -- Use `intrp` to recursively call the interpreter.
  let intrp := interp.stmt n in
  none
| (n + 1) env s h (stmt.while e p) :=
  let int := interp.stmt n in
  none
| (n + 1) env s h (stmt.seq p1 p2) :=
  do (s', h') ← interp.stmt n env s h p1,
     interp.stmt n env s' h' p2

def interp.program (ioi : io.interface) (fuel : nat) : prog → option (heap × val)
| (prog.mk funcs body ret) :=
  match interp.stmt ioi fuel funcs store.empty heap.empty body with
  | some (s', h') :=
    match interp.exp s' h' ret with
    | some v' := some (h', v')
    | none := none
    end
  | none := none
  end

end imp
