import ..to_python
import .syntax
import .has_to_string

namespace imp

def val.to_python : val → string
| (val.bool tt) := "True"
| (val.bool ff) := "False"
| (val.int i) := to_string i
| (val.str s) := repr s
| (val.addr i) := i.repr

open unary_op

def unary_op.to_python : unary_op → string
| neg := "-"
| not := "not"

open binary_op
def binary_op.to_python : binary_op → string
| add := "+"
| sub := "-"
| mul := "*"
| div := "/"
| mod := "%"
| eq := "=="
| lt := "<"
| le := "<="
| conj := "and"
| disj := "or"

def expr.to_python : expr → string
| (expr.val v) := v.to_python
| (expr.var x) := x
| (expr.unary_op op e1) := op.to_python ++ " " ++ e1.to_python
| (expr.binary_op op e1 e2) := e1.to_python ++ " " ++ op.to_python ++ " " ++ e2.to_python
| (expr.length e) := "len(" ++ e.to_python ++ ")"
| (expr.index e1 e2) := e1.to_python ++ "[" ++ e2.to_python ++ "]"

def stmt.to_python' : stmt → list string
| (stmt.nop) := "" :: []
| (stmt.set x e) := [to_string x ++ " = " ++ e.to_python]
| (stmt.alloc x e1 e2) := [to_string x ++ " = " ++ e1.to_python ++ " * [" ++ e2.to_python ++ "]"]
| (stmt.write x e1 e2) := [to_string x ++ "[" ++ e1.to_python ++ "]" ++ " = " ++ e2.to_python]
| (stmt.call x f es) :=
  [x ++ " = " ++ to_string f ++ (parens $ concat ", " $ es.map expr.to_python)]
| (stmt.seq s1 s2) := stmt.to_python' s1 ++ stmt.to_python' s2
| (stmt.ifelse e p1 stmt.nop) :=
  ["if " ++ e.to_python ++ ":"] ++ indent (stmt.to_python' p1)
| (stmt.ifelse e p1 p2) :=
  ["if " ++ e.to_python ++ ":"] ++
   indent (stmt.to_python' p1) ++ ["else:"] ++ indent (stmt.to_python' p2)
| (stmt.while e p1) :=
  ["while " ++ e.to_python ++ ":"] ++ indent (stmt.to_python' p1)

def stmt.to_python (s : stmt) :=
concat "\n" (stmt.to_python' s)

def func.to_python' : func → list string
| (func.mk name ps body ret) :=
  ("def " ++ name ++ parens (concat "," ps) ++ ":") :: indent (body.to_python' ++
  ["return " ++ ret.to_python])

def func.to_python (f : func) :=
concat "\n" (func.to_python' f)

instance func.has_to_python : has_to_python func :=
⟨ func.to_python ⟩

def prog.to_python' : prog → list string
| (prog.mk fs main ret) :=
  fs.map func.to_python ++ [main.to_python] ++ ["print " ++ "(" ++ ret.to_python ++ ")"]

def prog.to_python (p : prog) : string :=
  concat "\n\n" (prog.to_python' p)

instance prog.has_to_python : has_to_python prog :=
⟨ prog.to_python ⟩

end imp
