import .syntax
import ..print_util

namespace imp

def val.repr : val → string
 | (val.bool tt) := "(val.bool tt)"
 | (val.bool ff) := "(val.bool ff)"
 | (val.int i) := "(val.int " ++ repr i ++ ")"
 | (val.str s) := "(val.str " ++ repr s ++ ")"
 | (val.addr i) := "(val.addr " ++ i.repr ++ ")"

instance val.has_repr : has_repr val :=
⟨ val.repr ⟩

open unary_op

def unary_op.repr : unary_op → string
| neg := "neg"
| not := "not"

instance unary_op.has_repr : has_repr unary_op :=
⟨ unary_op.repr ⟩

open binary_op

def binary_op.repr : binary_op → string
| add := "add"
| sub := "sub"
| mul := "mul"
| div := "div"
| mod := "mod"
| eq := "eq"
| lt := "lt"
| le := "le"
| conj := "conj"
| disj := "disj"

instance binary_op.has_repr : has_repr binary_op :=
⟨ binary_op.repr ⟩

def expr.repr : expr → string
| (expr.val v) := "(expr.val " ++  v.repr ++ ")"
| (expr.var x) := "(expr.var " ++ x ++ ")"
| (expr.unary_op op e1) := "(expr.op1 " ++ op.repr ++ e1.repr  ++ ")"
| (expr.binary_op op e1 e2) := "(expr.op2 " ++ op.repr ++ " " ++ e1.repr ++ e2.repr ++ ")"
| (expr.length e) := "(expr.length " ++ e.repr ++ ")"
| (expr.index e1 e2) := "(expr.index " ++ e1.repr ++ "[" ++ e2.repr ++ "])"

instance expr.has_repr : has_repr expr :=
⟨ expr.repr ⟩

def stmt.repr_aux : stmt → list string
| (stmt.nop) := ["stmt.nop"]
| (stmt.set x e) := [parens $ "stmt.set " ++ repr x ++ " " ++ e.repr]
| (stmt.alloc x e1 e2) := [parens $ "stmt.alloc " ++ repr x ++ e1.repr ++ e2.repr]
| (stmt.write x e1 e2) := [parens $ "stmt.write " ++ e1.repr ++ " " ++ e2.repr]
| (stmt.call x fn args) := [x ++ fn ++ parens (concat "," (list.map expr.repr args)) ++ ";"]
| (stmt.ifelse e p1 p2) := ["(stmt.ifelse " ++ e.repr] ++ stmt.repr_aux p1 ++ stmt.repr_aux p2 ++ [")"]
| (stmt.while g p1) := ["(stmt.while " ++ g.repr] ++ stmt.repr_aux p1 ++ stmt.repr_aux p1 ++ [")"]
| (stmt.seq s1 s2) := ["stmt.seq "] ++ stmt.repr_aux s1 ++ stmt.repr_aux s2

def stmt.repr (s : stmt) :=
concat "\n" (stmt.repr_aux s)

instance stmt.has_repr : has_repr stmt :=
⟨ stmt.repr ⟩

def func.repr : func → string
| (func.mk name ps body ret) :=
  parens $ "func.mk " ++ repr name ++ squares (string.join $ list.intersperse "," $ ps.map repr) ++
   parens (concat "\n" $ indent $ body.repr_aux ++ [ "ret " ++ repr ret ++ ";"])

instance func.has_repr : has_repr func :=
⟨ func.repr ⟩

def prog.repr : prog → string
| (prog.mk funcs main ret) :=
  parens $ "prog.mk " ++ (concat "\n\n" $ list.map func.repr funcs) ++ repr main ++ repr ret

instance prog.has_repr : has_repr prog :=
⟨ prog.repr ⟩

end imp
