import struct_tact
import ..imp
import .eval

/- Miscellanous helper lemmas. -/

namespace imp

lemma locate_inv :
  forall env x name params body ret,
    locate env x = some (func.mk name params body ret) →
    x = name :=
begin
    induction_on env ; intros,
    { cases a },
    { cases a,
      by_cases (x = a),
      { simp [locate, *] at a_2,
        cases a_2,
        assumption
      },
      { simp [locate, h] at a_2,
          apply ih_1,
          assumption,
      }
    }
end

/- Evaluation facts -/

lemma eval_unary_op_det :
  forall op v v1 v2,
    eval.unary_op op v v1 →
    eval.unary_op op v v2 →
    v1 = v2 :=
begin
  intros; cases v ; cases a,
  { cases a_1, reflexivity },
  { cases a_1, reflexivity }
end

lemma eval_binary_op_det :
  forall op vL vR v1 v2,
    eval.binary_op op vL vR v1 →
    eval.binary_op op vL vR v2 →
    v1 = v2 :=
begin
  intros,
  cases a ; cases a_1 ; trivial,
end

lemma eval_expr_det :
  forall s h e v1 v2,
    eval.expr s h e v1 ->
    eval.expr s h e v2 ->
    v1 = v2 :=
begin
  intros,
  revert v2,
  induction a; intros,
  case eval.expr.val {
    cases a_1,
    reflexivity
  },
  case eval.expr.var {
    cases a_1,
    simp * at *,
    cases a_2,
    reflexivity
  },
  case eval.expr.unary_op {
    cases a_1_1,
    have x := ih_1 v_1 a_2,
    rw x at a_1,
    apply eval_unary_op_det ; assumption
  },
  case eval.expr.binary_op {
    cases a_1_1,
    apply eval_binary_op_det,
    apply a_2,
    specialize (ih_1 _ a_3),
    specialize (ih_2 _ a_4),
    simp * at *,
  },
  case eval.expr.length_a {
    cases a_1_1,
    { specialize ih_1 _ a_4, cases ih_1,
      rw a_5 at a_2, cases a_2, trivial },
    { specialize ih_1 _ a_3, cases ih_1 },
  },
  case eval.expr.length_s {
    cases a_1_1,
    { specialize ih_1 _ a_3, cases ih_1 },
    { specialize ih_1 _ a_2, cases ih_1,
      rw a_3 at a_1, cases a_1, trivial }
  },
  case eval.expr.index_s {
    cases a_1_1,
    specialize ih_1 _ a_5,
    cases ih_1,
    specialize ih_1 _ a_4,
    cases ih_1,
    specialize ih_2 _ a_5,
    cases ih_2,
    rw a_7 at a_3,
    cases a_3,
    reflexivity,
  },
  case eval.expr.index_a {
    cases a_1_1,
    specialize ih_1 _ a_7,
    cases ih_1,
    specialize ih_2 _ a_8,
    cases ih_2,
    rw a_11 at a_5,
    cases a_5,
    reflexivity,
    specialize ih_1 _ a_6,
    cases ih_1,
  }
end

lemma eval_e_len_a_inv :
  forall s h e x a l,
    eval.expr s h (expr.length e) x →
    eval.expr s h e (val.addr a) →
    heap.read h a = some (val.int l) →
    x = val.int l :=
begin
  intros,
  cases a_1,
  { apply eval_expr_det,
    assumption,
    have eq : val.addr a = val.addr a_4,
    apply eval_expr_det ; assumption,
    cases eq,
    rw a_3 at a_6,
    cases a_6,
    assumption
  },
  { have contra : val.addr a = val.str cs,
    apply eval_expr_det ; assumption,
    contradiction,
  }
end

end imp
