#!/usr/bin/env bash

TDIR="$(cd "$(dirname "${BASH_SOURCE[0]}")" && pwd)"
LEAN=`which lean`
MAIN="${TDIR}/../../src/main.lean"

if ! [ -f "$LEAN" ]; then
  echo "ERROR: could not find $LEAN, please build first"
  exit 1
fi

if [ $# -ne 2 ]; then
  echo "Usage: $0 MININT MAXINT"
  exit 1
fi

MININT=$1
MAXINT=$2

function go {
  prog="$1"
  shift

  input="$(mktemp TESTIMP.XXXXX)"
  for a in "$@"; do
    echo "$a" >> "$input"
  done

  for a in "$@"; do
    echo "$a" >> "$input"
  done

  if cat "$input" | "$LEAN" --run "$MAIN" --test "$prog" > /dev/null 2>&1 ; then
    rm "$input"
  else
    echo "FAIL $prog on:"
    cat "$input"
    exit 1
  fi
}

function bloop {
  for x in "False" "True"; do
    $@ "$x"
  done
}

function iloop {
  for x in $(seq $MININT $MAXINT); do
    $@ "$x"
  done
}

function sloop {
  for x in "\"\"" "a" "ab" "abc"; do
    $@ "$x"
  done
}

echo "bounded exhaustive testing... (this takes several minutes)"

pushd "$TDIR" > /dev/null
bloop go not.imp
iloop go neg.imp

iloop iloop go add.imp
sloop sloop go sadd.imp
iloop iloop go sub.imp
iloop iloop go mul.imp
iloop iloop go div.imp
iloop iloop go mod.imp

iloop iloop go eq-i.imp
bloop bloop go eq-b.imp
iloop iloop go le.imp
iloop iloop go lt.imp

bloop bloop go conj.imp
bloop bloop go disj.imp
